from typing import List, Optional

import pandas as pd
from rdflib import Graph

from pandasaurus.curie_validator import CurieValidator
from pandasaurus.graph.graph_generator import GraphGenerator
from pandasaurus.resources.term import Term
from pandasaurus.slim_manager import SlimManager
from pandasaurus.utils.pandasaurus_exceptions import InvalidTerm, ObsoletedTerm
from pandasaurus.utils.query_utils import chunks, run_sparql_query
from pandasaurus.utils.sparql_queries import (
    get_ancestor_enrichment_query,
    get_contextual_enrichment_query,
    get_full_enrichment_query,
    get_most_specific_objects_query,
    get_most_specific_subjects_query,
    get_simple_enrichment_query,
    get_synonym_query,
)


class Query:
    """Query class is responsible for generating a pandas DataFrame that enriches the seed list with synonyms and
    all inferred subClassOf relationships between terms in the seed. It also allows queries over the DataFrame.

    Attributes:
        enriched_df: DataFrame that is enriched with synonyms and inferred relationships between terms in
        the seed. It will be used in further filtered queries.

    """

    def __init__(
        self,
        seed_list: List[str],
        enrichment_property_list: Optional[List[str]] = None,
        force_fail: bool = False,
    ):
        """A Query object is initialised by passing a list of seed terms (where each term is a CURIE string,
        e.g. CL:0000001; all OBO standard CURIESs are recognised). It generates a pandas DataFrame that enriches the
        seed list with synonyms and all inferred subClassOf relationships by default, and it supports other
        relationships between terms in the seed. Additional methods allow enrichment with terms outside the seed
        from slims or specified by a semantic context.

        Args:
            seed_list: A list of seed terms where each term is a CURIE string
            enrichment_property_list: Optional list of property IRIs to extend enrichment queries.

        """
        # Might be unnecessary
        self._seed_list = seed_list
        self._enrichment_property_list = enrichment_property_list if enrichment_property_list else ["rdfs:subClassOf"]
        self._term_list: List[Term] = CurieValidator.construct_term_list(seed_list)
        self.enriched_df = pd.DataFrame()
        self.graph_df = pd.DataFrame()
        self.graph = Graph()
        # Validation and reporting
        try:
            CurieValidator.get_validation_report(self._term_list)
        except InvalidTerm as e:
            print(e.message)
            if force_fail:
                raise ValueError("Check your seed list! It contains invalid terms")
        except ObsoletedTerm as e:
            print(e.message)
            if force_fail:
                raise ValueError(
                    "Check your seed list! It contains obsoleted terms. Use update_obsoleted_terms "
                    "method to update all obsoleted term"
                )

    def simple_enrichment(self) -> pd.DataFrame:
        """Returns a DataFrame that is enriched with synonyms and inferred relationships between terms in the seed.
        Subject and object terms are members of the seed terms.

        Returns:
             Enriched DataFrame

        """
        source_list = [term.get_iri() for term in self._term_list]
        object_list = source_list
        query_string = get_simple_enrichment_query(source_list, object_list, self._enrichment_property_list)
        self.enriched_df = (
            pd.DataFrame(
                [res for res in run_sparql_query(query_string)],
                columns=["s", "s_label", "p", "o", "o_label"],
            )
            .sort_values("s")
            .reset_index(drop=True)
        )
        self._generate_enrichment_graph(object_list)

        return self.enriched_df

    def minimal_slim_enrichment(self, slim_list: List[str]) -> pd.DataFrame:
        """Returns a DataFrame that is enriched with synonyms and inferred relationships between terms in the seed list
        and in an extended seed list. The extended seed list consists of terms from the seed list and terms from
        given slim lists, classes tagged with some specified ‘subset’ axiom.

        Args:
            slim_list: List 'subset' tags that consists of classes tagged with some specified ‘subset’ axiom

        Returns:
            Enriched DataFrame

        """
        source_list = [term.get_iri() for term in self._term_list]
        object_list = list(set(source_list + SlimManager.get_slim_members(slim_list)))
        s_result = []
        for chunk in chunks(object_list, 90):
            s_result.extend(
                [
                    res
                    for res in run_sparql_query(
                        get_simple_enrichment_query(source_list, chunk, self._enrichment_property_list)
                    )
                ]
            )
        self.enriched_df = (
            pd.DataFrame(s_result, columns=["s", "s_label", "p", "o", "o_label"])
            .sort_values("s")
            .reset_index(drop=True)
        )
        self._generate_enrichment_graph(object_list)

        return self.enriched_df

    def full_slim_enrichment(self, slim_list: List[str]) -> pd.DataFrame:
        """Returns a DataFrame that is enriched with synonyms and inferred relationships between terms in the seed list
        and in an extended seed list. The extended seed list consists of terms from the seed list and terms from
        given slim lists, classes tagged with some specified ‘subset’ axiom, with inferred terms via transitive
        subClassOf queries.

        Args:
             slim_list: List 'subset' tags that consists of classes tagged with some specified ‘subset’ axiom

         Returns:
             Enriched DataFrame

        """
        source_list = [term.get_iri() for term in self._term_list]
        object_list = list(set(source_list + SlimManager.get_slim_members(slim_list)))
        s_result = []
        for chunk in chunks(object_list, 90):
            s_result.extend([res for res in run_sparql_query(get_full_enrichment_query(source_list, chunk))])

        self.enriched_df = (
            pd.DataFrame(s_result, columns=["s", "s_label", "p", "x", "x_label"])
            .rename(columns={"x": "o", "x_label": "o_label"})
            .fillna({"p": "rdfs:subClassOf"})
            .sort_values("s")
            .reset_index(drop=True)
        )
        self._generate_enrichment_graph(object_list)

        return self.enriched_df

    def contextual_slim_enrichment(self, context: List[str]) -> pd.DataFrame:
        """Returns a DataFrame that is enriched with synonyms and inferred relationships between terms in the seed list
        and in an extended seed list. The extended seed list consists of terms from the seed list and all terms
        satisfied by some set of existential restrictions in the ubergraph (e.g. part_of some 'kidney').

        Args:
            context: Organ/tissue/multicellular anatomical structure list to determine the redundant graph via
            existential restrictions. It must be a valid CURIE.

        Returns:
            Enriched DataFrame

        """
        # TODO add a curie checking mechanism for context list
        query_string = get_contextual_enrichment_query(context)
        source_list = [term.get_iri() for term in self._term_list]
        object_list = list(set(source_list + [res.get("term") for res in run_sparql_query(query_string)]))
        s_result = []
        for chunk in chunks(object_list, 90):
            s_result.extend(
                [
                    res
                    for res in run_sparql_query(
                        get_simple_enrichment_query(source_list, chunk, self._enrichment_property_list)
                    )
                ]
            )

        self.enriched_df = (
            pd.DataFrame(s_result, columns=["s", "s_label", "p", "o", "o_label"])
            .sort_values("s")
            .reset_index(drop=True)
        )
        self._generate_enrichment_graph(object_list)

        return self.enriched_df

    def ancestor_enrichment(self, step_count: str) -> pd.DataFrame:
        """
        Perform ancestor enrichment analysis with a specified number of hops.

        Args:
            step_count (str): The number of hops to consider when enriching terms.

        Returns:
            pd.DataFrame: A DataFrame containing enriched terms and associated information.

        This method conducts an ancestor enrichment analysis on a set of seed terms,
        considering the specified number of hops in the ontology graph. The analysis
        retrieves terms that are ancestors of the seed terms within the specified
        number of hops and compiles the results into a DataFrame.

        The `step_count` parameter controls the depth of the analysis. A smaller
        `step_count` limits the analysis to immediate ancestors, while a larger value
        includes more distant ancestors.

        """
        source_list = [term.get_iri() for term in self._term_list]
        query_string = get_ancestor_enrichment_query(source_list, step_count)
        object_list = list(set(uri for res in run_sparql_query(query_string) for uri in res.values()))
        s_result = []
        for chunk in chunks(object_list, 90):
            s_result.extend(
                [
                    res
                    for res in run_sparql_query(
                        get_simple_enrichment_query(source_list, chunk, self._enrichment_property_list)
                    )
                ]
            )

        self.enriched_df = (
            pd.DataFrame(s_result, columns=["s", "s_label", "p", "o", "o_label"])
            .sort_values("s")
            .reset_index(drop=True)
        )
        self._generate_enrichment_graph(object_list)

        return self.enriched_df

    def parent_enrichment(self):
        """
        Perform parent enrichment analysis.

        This method is a convenience wrapper around the `ancestor_enrichment` method,
        specifically designed to perform parent enrichment analysis. Parent enrichment
        analysis considers only immediate parent terms of the seed terms in the ontology
        graph (i.e., one-hop ancestors).

        Returns:
            pd.DataFrame: A DataFrame containing enriched parent terms and associated
            information.

        This method simplifies the process of conducting parent enrichment analysis by
        calling the `ancestor_enrichment` method with a `step_count` of 1, which limits
        the analysis to immediate parent terms of the seed terms.

        """
        self.ancestor_enrichment(1)

    def synonym_lookup(self) -> pd.DataFrame:
        """

        Returns:
            A DataFrame containing labels and synonyms of the terms from the seed list.

        """
        label_df = pd.DataFrame(
            {term.get_iri(): term.get_label() for term in self._term_list}.items(), columns=["ID", "label"]
        )

        synonym_query_results = run_sparql_query(get_synonym_query(label_df["ID"].tolist()))
        synonym_df = (
            pd.DataFrame([res for res in synonym_query_results if any("synonym" in key for key in res.keys())])
            .melt(id_vars="s", var_name="type", value_name="name")
            .rename(columns={"s": "ID"})
            .dropna(subset=["name"])[["ID", "name", "type"]]
        )

        # Merging two df
        result_df = (
            pd.merge(synonym_df, label_df, on="ID", how="left")
            .sort_values("ID")
            .reset_index(drop=True)[["ID", "label", "name", "type"]]
        )

        return result_df

    def get_most_specific_objects(self, predicate: str, ontology: str):
        """

        Args:
            predicate: Relationship that wanted to be explored
            ontology: PURL of obo ontologies in Ubergraph.

        Examples:
            Example Ontology PURLs:
                - http://purl.obolibrary.org/obo/cl.owl \n
                - http://purl.obolibrary.org/obo/uberon.owl

        Returns:

        """
        subject_list = [term.get_iri() for term in self._term_list]
        query_string = get_most_specific_objects_query(subject_list, predicate, ontology)
        return (
            pd.DataFrame(
                [res for res in run_sparql_query(query_string)],
                columns=["s", "s_label", "p", "o", "o_label"],
            )
            .sort_values("s")
            .reset_index(drop=True)
        )

    def get_most_specific_subjects(self, predicate: str, ontology: str):
        """

        Args:
            predicate: Relationship that wanted to be explored
            ontology: PURL of obo ontologies in Ubergraph.

        Examples:
            Example Ontology PURLs:
                - http://purl.obolibrary.org/obo/cl.owl \n
                - http://purl.obolibrary.org/obo/uberon.owl

        Returns:

        """
        object_list = [term.get_iri() for term in self._term_list]
        query_string = get_most_specific_subjects_query(object_list, predicate, ontology)
        return (
            pd.DataFrame(
                [res for res in run_sparql_query(query_string)],
                columns=["s", "s_label", "p", "o", "o_label"],
            )
            .sort_values("s")
            .reset_index(drop=True)
        )

    def query(self, column_name: str, query_term: str) -> pd.DataFrame:
        """Returns filtered dataframe via join on column to subject of enriched_df, looking up of object name or
        synonym via query of name_lookup.

        Args:
            column_name: Column name
            query_term: Object label or synonym

        Returns:
            Filtered DataFrame

        """
        # TODO Add missing implementation
        raise NotImplementedError

    def update_obsoleted_terms(self):
        """Replaces all obsoleted terms in the term list with the new term that obsoletes them."""
        [getattr(term, "update_obsoleted_term")() for term in self._term_list]

    def mirror_enrichment_for_graph_generation(self, term_list: List[str]):
        # TODO definitely need a refactoring later on
        s_result = []
        for s_chunk in chunks(term_list, 45):
            for o_chunk in chunks(term_list, 45):
                s_result.extend(
                    [
                        res
                        for res in run_sparql_query(
                            get_simple_enrichment_query(s_chunk, o_chunk, self._enrichment_property_list)
                        )
                    ]
                )
        self.graph_df = (
            pd.DataFrame(s_result, columns=["s", "s_label", "p", "o", "o_label"])
            .sort_values("s")
            .reset_index(drop=True)
        )

    def _generate_enrichment_graph(self, object_list):
        self.mirror_enrichment_for_graph_generation(object_list)
        self.graph = GraphGenerator.generate_enrichment_graph(self.graph_df)
        self.graph = GraphGenerator.apply_transitive_reduction(self.graph, self.enriched_df["p"].unique().tolist())



from typing import List

import networkx as nx
import pandas as pd
from rdflib import OWL, RDF, RDFS, Graph, Literal, Namespace, URIRef
from rdflib.plugins.sparql import prepareQuery

from pandasaurus.graph.graph_generator_utils import (
    add_edge,
    add_outgoing_edges_to_subgraph,
)
from pandasaurus.utils.logging_config import configure_logger

# Set up logger
logger = configure_logger()


class GraphGenerator:
    @staticmethod
    def generate_enrichment_graph(enriched_df: pd.DataFrame) -> Graph:
        """
        Generates an RDF graph representing enrichment relationships from an enriched DataFrame.

        Args:
            enriched_df (pd.DataFrame): A DataFrame containing enrichment relationships.
                It should have columns 's', 's_label', 'o', and 'o_label' representing subjects,
                subject labels, objects, and object labels respectively.

        Returns:
            Graph: An RDF graph representing the enrichment relationships.
                Each subject is linked to its corresponding object using the 'subClassOf' relationship,
                and labels are associated with subjects and objects using the 'label' relationship.
        """
        graph = Graph()
        cl_namespace = Namespace("http://purl.obolibrary.org/obo/CL_")
        for _, row in enriched_df.iterrows():
            s = cl_namespace[row["s"].split(":")[-1]]
            o = cl_namespace[row["o"].split(":")[-1]]
            graph.add((s, RDFS.label, Literal(row["s_label"])))
            graph.add((s, RDF.type, OWL.Class))
            graph.add((o, RDFS.label, Literal(row["o_label"])))
            graph.add((o, RDF.type, OWL.Class))
            graph.add((s, RDFS.subClassOf, o))
        return graph

    @staticmethod
    def apply_transitive_reduction(graph: Graph, predicate_list: List[str]) -> Graph:
        """
        Applies transitive reduction to a given RDF graph using specified predicates.

        Args:
            graph (Graph): The RDF graph to which transitive reduction is applied.
            predicate_list (List[str]): A list of predicates (URIs or 'rdfs:subClassOf')
                to which transitive reduction will be applied.

        Returns:
            Graph: The RDF graph after applying transitive reduction to the specified predicates.

        Note:
            - The method applies transitive reduction to the specified predicates by adding only the necessary edges
              to create a directed acyclic graph.
            - Redundant triples are removed from the graph using transitive reduction results.
            - The 'predicate_list' should contain valid predicates existing in the graph.
            - For 'predicate_list', you can use either full URIs or 'rdfs:subClassOf' for the RDF schema 'subClassOf'
              relationship.
        """
        invalid_predicates = []
        # TODO We need a better way to handle the queries, and decide the format we accept in the predicate list
        ask_query = prepareQuery("SELECT ?s ?p WHERE { ?s ?p ?o }")
        for predicate in predicate_list:
            predicate_ = RDFS.subClassOf if predicate == "rdfs:subClassOf" else URIRef(predicate)
            if predicate and not graph.query(ask_query, initBindings={"p": predicate_}, initNs={"rdfs": RDFS}):
                invalid_predicates.append(predicate)
                continue

            subgraph = add_outgoing_edges_to_subgraph(graph, predicate_)

            nx_graph = nx.DiGraph()
            for s, p, o in subgraph:
                if isinstance(o, URIRef) and p != RDF.type:
                    add_edge(nx_graph, s, predicate, o)

            # Apply transitive reduction to remove redundancy
            transitive_reduction_graph = nx.transitive_reduction(nx_graph)
            transitive_reduction_graph.add_edges_from(
                (u, v, nx_graph.edges[u, v]) for u, v in transitive_reduction_graph.edges
            )
            # Remove redundant triples using nx graph
            edge_diff = list(set(nx_graph.edges) - set(transitive_reduction_graph.edges))
            for edge in edge_diff:
                graph.remove((URIRef(edge[0]), predicate_, URIRef(edge[1])))
            # TODO Temporarily disabling this log message
            # logger.info(f"Transitive reduction has been applied on {predicate} for graph generation.")

        if invalid_predicates:
            error_msg = (
                f"The predicate '{invalid_predicates[0]}' does not exist in the graph"
                if len(invalid_predicates) == 1
                else f"The predicates {' ,'.join(invalid_predicates)} do not exist in the graph"
            )
            logger.error(error_msg)

        return graph



from typing import List

from pandasaurus.resources.term import Term


class InvalidTerm(Exception):
    def __init__(self, term_list: List[Term]):
        """
        Exception raised for invalid ontology terms.

        Args:
            term_list (List[Term]): A list of invalid Term objects.
        """
        self.message = f"The following terms are invalid: {', '.join([term.get_iri() for term in term_list])}"
        super().__init__(self.message)


class ObsoletedTerm(Exception):
    def __init__(self, term_list: List[Term]):
        """
        Exception raised for obsoleted ontology terms.

        Args:
            term_list (List[Term]): A list of obsoleted Term objects.
        """
        self.message = (
            f"The following terms are obsoleted: {', '.join([term.get_iri() for term in term_list])}, "
            f"and replaced by following terms: {', '.join([term.get_new_iri() for term in term_list])}. "
            f"Please consider using the new terms"
        )
        super().__init__(self.message)


class InvalidOntology(Exception):
    def __init__(self, input_ontology: str, ontology_list: List[str]):
        """
        Exception raised for invalid ontology names.

        Args:
            input_ontology (str): The name of the invalid ontology.
            ontology_list (List[str]): A list of valid ontology names.
        """
        self.message = (
            f"The '{input_ontology}' ontology is invalid. \n"
            f"Please use one of the following ontologies: \n{', '.join(ontology_list)}"
        )
        super().__init__(self.message)



from rdflib import Graph


def add_edge(nx_graph, subject, predicate, obj):
    edge_data = {"label": str(predicate).split("#")[-1] if "#" in predicate else str(predicate).split("/")[-1]}
    nx_graph.add_edge(
        str(subject),
        str(obj),
        **edge_data,
    )


def add_outgoing_edges_to_subgraph(graph, predicate_uri=None):
    """
    Add all outgoing edges of a node in the graph to the subgraph.

    Parameters:
        graph (Graph): The RDF graph containing the triples.
        predicate_uri (URIRef or None): The predicate to filter triples (optional).

    Returns:
        rdflib.Graph: The subgraph containing the outgoing edges of the nodes.
    """
    subgraph = Graph()
    for s, p, o in graph.triples((None, predicate_uri, None)):
        subgraph.add((s, p, o))

    return subgraph



from typing import Dict, List

from pandasaurus.resources.slim import Slim
from pandasaurus.utils.pandasaurus_exceptions import InvalidOntology
from pandasaurus.utils.query_utils import run_sparql_query
from pandasaurus.utils.sparql_queries import (
    get_ontology_list_query,
    get_slim_list_query,
    get_slim_members_query,
)


class SlimManager:
    """SlimManager responsible for slim operations such as finding available slim in given ontologies and showing slim
    content.
    """

    # Might not be needed
    @staticmethod
    def get_slim_list(ontology: str) -> List[Dict[str, str]]:
        """Returns name and definition of available slims in given ontology.

        Args:
            ontology: Ontology name

        Returns:
            Slim names and definitions list

        """
        ontology_list = SlimManager._get_ontology_list()
        if ontology in ontology_list:
            slim_list: Dict[str, Slim] = dict()
            result = run_sparql_query(get_slim_list_query(ontology))
            for res in result:
                slim_list.update({res.get("label"): Slim(name=res.get("label"), description=res.get("comment"))})
            return [{"name": slim.get_name(), "description": slim.get_description()} for slim in slim_list.values()]
        raise InvalidOntology(ontology, ontology_list)

    @staticmethod
    def get_slim_members(slim_list: List[str]) -> List[str]:
        """Lists names and IDs of all terms in slim.

        Args:
            slim_list: A list of slim names

        Returns:
            Term IRIs of the slim members

        """
        return [
            term.get("term") for slim_name in slim_list for term in run_sparql_query(get_slim_members_query(slim_name))
        ]

    @staticmethod
    def _get_ontology_list():
        return [row.get("title") for row in run_sparql_query(get_ontology_list_query())]



from abc import abstractmethod
from typing import Dict, List, Union, Any, Optional

from pandasaurus.resources.term import Term
from pandasaurus.utils.pandasaurus_exceptions import InvalidTerm, ObsoletedTerm
from pandasaurus.utils.query_utils import run_sparql_query
from pandasaurus.utils.sparql_queries import get_label_query, get_replaced_by_query


class CurieValidator:
    """CurieValidator is responsible for validating CURIE prefixes and CURIEs of slim terms. It also suggests
    replacements for obsoleted slim terms.
    """

    @staticmethod
    @abstractmethod
    def validate_curie_prefixes(curie_list: List[str]) -> Dict[str, bool]:
        """Reports whether the CURIE prefixes are valid or not.

        Args:
            curie_list: List of CURIEs

        Returns:
            True or False status of the prefix validation for each term

        """
        # Is there anyway to validate prefixes via Ubergraph or are we going to validate them via a config file? OAK?
        # Do we still need this? https://github.com/INCATools/PandaSaurus/issues/1#issuecomment-1527753655
        # TODO Add missing implementation
        raise NotImplementedError

    @staticmethod
    def validate_curie_list(curie_list: List[str]) -> dict[str, dict[str, Union[Optional[bool], Any]]]:
        """Reports whether the CURIEs are valid or not.

        Args:
            curie_list: List of CURIEs

        Examples:
            | An example output that shows 2 valid and 1 invalid CURIEs:
            | {'CL:0002681': {'label': 'kidney cortical cell', 'valid': True},
            |  'CL:0002518': {'label': 'kidney epithelial cell', 'valid': True},
            |  'CL:1234567': {'label': None, 'valid': False}}

        Returns:
            True or False status of the CURIE validation for each term

        """
        query_string = get_label_query(curie_list)
        result_dict = dict([(r.get("term"), r.get("label")) for r in run_sparql_query(query_string)])
        return {
            curie: {
                "label": result_dict.get(curie) if curie in result_dict else None,
                "valid": True if curie in result_dict else False,
            }
            for curie in curie_list
        }

    @staticmethod
    def find_obsolete_terms(curie_list: List[str]) -> Dict:
        """Returns obsoleted terms in the curie_list and label and IRI for of the term that is replaced the obsoleted
        terms.

        Args:
            curie_list: List of CURIEs

        Examples:
            | An example output that shows 1 obsoleted CURIE:
            | {'CL:0011107': {'term': 'CL:0011107', 'label': 'obsolete Muller cell', depr_status': 'true',
            | 'new_term': 'CL:0000636', new_term_label': 'Mueller cell'}

        Returns:
            True or False status of the term for each term

        """
        query_string = get_replaced_by_query(curie_list)
        result_dict = dict([(r.get("term"), r) for r in run_sparql_query(query_string)])
        return result_dict

    @staticmethod
    @abstractmethod
    def find_obsolete_term_replacement(curie_list: Dict[str, str]) -> Dict[str, str]:
        """Suggests terms for each obsoleted terms in the curie_list.

        Args:
            curie_list: List of CURIEs

        Returns:
            List of suggested term

        """
        # TODO Add missing implementation
        # We probably don't need this method
        raise NotImplementedError

    @staticmethod
    def get_validation_report(term_list: List[Term]):
        """Returns validation report which includes invalid and obsoleted terms.

        Args:
            term_list: A list of seed terms where each term is a Term object

        """
        invalid_terms: List[Term] = []
        obsoleted_terms: List[Term] = []
        for term in term_list:
            if not term.get_is_valid():
                invalid_terms.append(term)
            if term.get_is_obsoleted():
                obsoleted_terms.append(term)

        if invalid_terms:
            raise InvalidTerm(invalid_terms)
        if obsoleted_terms:
            raise ObsoletedTerm(obsoleted_terms)

    @staticmethod
    def construct_term_list(seed_list) -> List[Term]:
        """Returns list of Term objects after running validate_curie_list and find_obsolete_terms methods.

        Args:
            seed_list: A list of seed terms where each term is a CURIE string

        Returns:
            List of Term objects

        """
        term_validation = CurieValidator.validate_curie_list(seed_list)
        term_obsoletion = CurieValidator.find_obsolete_terms(seed_list)
        term_list: List[Term] = list()
        for seed in seed_list:
            term = Term(
                term_validation.get(seed).get("label"),
                seed,
                term_validation.get(seed).get("valid"),
                term_obsoletion.get(seed).get("new_term_label") if seed in term_obsoletion else None,
                term_obsoletion.get(seed).get("new_term") if seed in term_obsoletion else None,
            )
            term_list.append(term)
        return term_list



from typing import Iterator

from oaklib.implementations import UbergraphImplementation

oi = UbergraphImplementation()


def run_sparql_query(query: str) -> Iterator:
    return oi.query(query=query, prefixes=get_prefixes(query, oi.prefix_map().keys()))


def chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def get_prefixes(text, prefix_map):
    _prefixes = []
    for prefix in prefix_map:
        if prefix + ":" in text:
            _prefixes.append(prefix)

    return _prefixes



from typing import Optional


class Term:
    """Represents ontology terms."""

    def __init__(
        self,
        label: str,
        iri: str,
        is_valid: bool,
        new_label: Optional[str] = None,
        new_iri: Optional[str] = None,
    ):
        self.__label = label
        self.__iri = iri
        self.__is_valid = is_valid
        self.__new_label = new_label
        self.__new_iri = new_iri
        self.__is_obsolete: bool = True if new_label and new_iri else False

    def get_label(self) -> str:
        """Returns term label.

        Returns:
            Term label

        """
        return self.__label

    def get_iri(self) -> str:
        """Returns term IRI.

        Returns:
            Term IRI

        """
        return self.__iri

    def get_is_valid(self) -> bool:
        """Returns term validation status.

        Returns:
            True if term is valid False otherwise

        """
        return self.__is_valid

    def get_new_label(self) -> str:
        """Returns new term label of obsoleted term.

        Returns:
            New term label

        """
        return self.__new_label

    def get_new_iri(self) -> str:
        """Returns new term IRI of obsoleted term.

        Returns:
            New term IRI

        """
        return self.__new_iri

    def get_is_obsoleted(self) -> str:
        """Returns term obsoletion status.

        Returns:
            True if term is obsoleted False otherwise

        """
        return self.__is_obsolete

    def update_obsoleted_term(self):
        """Updates term label and IRI if the term is obsoleted, and changes its obsoletion status."""
        if self.__is_obsolete:
            self.__label = self.__new_label
            self.__iri = self.__new_iri
            self.__is_obsolete = False

    def __eq__(self, other):
        if isinstance(other, Term):
            return (
                self.__label == other.__label
                and self.__iri == other.__iri
                and self.__is_valid == other.__is_valid
                and self.__is_obsolete == other.__is_obsolete
                and (self.__new_label == other.__new_label if (self.__is_obsolete and other.__is_obsolete) else True)
                and (self.__new_iri == other.__new_iri if (self.__is_obsolete and other.__is_obsolete) else True)
            )
        return False

    def __str__(self):
        msg = f"IRI: {self.__iri}, Label: {self.__label}, Valid: {self.__is_valid}"
        if self.__is_valid:
            msg += f", Obsoleted: {self.__is_obsolete}"
        if self.__is_obsolete:
            msg += f", New term label: {self.__new_label}, New term IRI: {self.__new_iri}"
        return msg



from typing import List


# TODO might need to update the enrichment query method names
def get_simple_enrichment_query(s_iri_list: List[str], o_iri_list: List[str], property_list: List[str]) -> str:
    return (
        f"SELECT ?s ?s_label ?p ?o ?o_label WHERE {{ GRAPH <http://reasoner.renci.org/redundant> {{"
        f"VALUES ?s {{ {' '.join(s_iri_list)} }} VALUES ?o {{ {' '.join(o_iri_list)} }} "
        f"VALUES ?p {{ {' '.join(property_list)} }}"
        f"?s ?p ?o. }} ?s rdfs:label ?s_label. ?o rdfs:label ?o_label. FILTER(?s != ?o)}}# LIMIT"
    )


def get_minimal_enrichment_query(seed_list: List[str]) -> str:
    # TODO Add missing implementation. Might not be needed
    raise NotImplementedError


def get_full_enrichment_query(s_iri_list: List[str], o_iri_list: List[str]) -> str:
    return (
        f"SELECT DISTINCT ?s ?s_label ?p ?x ?x_label ?o ?o_label WHERE {{ GRAPH "
        f"<http://reasoner.renci.org/nonredundant> {{ VALUES ?s {{ {' '.join(s_iri_list)} }} VALUES ?o {{ "
        f" {' '.join(o_iri_list)} }}  ?s rdfs:subClassOf* ?x. ?x rdfs:subClassOf* ?o. FILTER(?s != ?x)}} ?s  "
        f"rdfs:label ?s_label. ?x rdfs:label  ?x_label. ?o rdfs:label ?o_label }}# LIMIT"
    )


def get_contextual_enrichment_query(context_list: List[str]) -> str:
    # BFO:0000050 , part of
    return (
        f"SELECT * WHERE {{ VALUES ?context {{{' '.join(context_list)} }} ?term BFO:0000050 ?context. "
        f"?term rdfs:subClassOf CL:0000000. ?term rdfs:label ?label }}# LIMIT"
    )


def get_curie_prefix_validation_query(seed_list: List[str]) -> str:
    # TODO Add missing implementation. Might not be needed
    raise NotImplementedError


def get_label_query(term_iri_list: List[str]) -> str:
    """Used for CURIE validation"""
    return (
        f"SELECT ?term ?label where {{ ?term rdf:type owl:Class. OPTIONAL {{?term rdfs:label ?label.}} "
        f"VALUES ?term {{ {' '.join(term_iri_list)}}} }}# LIMIT"
    )


def get_ancestor_enrichment_query(seed_list: List[str], step_count) -> str:
    query = "SELECT * WHERE { GRAPH <http://reasoner.renci.org/nonredundant> "
    query += f"{{ VALUES ?s {{{' '.join(seed_list)} }} "
    query += "?s rdfs:subClassOf ?o0. "

    defined_by = (
        "GRAPH <http://reasoner.renci.org/ontology> { ?o0 rdfs:isDefinedBy " "<http://purl.obolibrary.org/obo/cl.owl>. "
    )

    # Build the nested OPTIONAL blocks for the specified number of steps
    for step in range(step_count):
        if step < step_count - 1:
            query += f"OPTIONAL {{ ?o{step} rdfs:subClassOf ?o{step + 1}.}} "
            defined_by += f"?o{step + 1} rdfs:isDefinedBy <http://purl.obolibrary.org/obo/cl.owl>. "

    query += "} "
    query += defined_by
    query += "}} #LIMIT"

    return query


def get_synonym_query(term_iri_list: List[str]) -> str:
    return (
        f"SELECT * WHERE {{VALUES ?s {{ {' '.join(term_iri_list)} }}"
        f"{{ OPTIONAL {{ ?s oio:hasNarrowSynonym ?narrow_synonym }} }} "
        f"UNION {{ OPTIONAL {{ ?s oio:hasExactSynonym ?exact_synonym }} }} "
        f"UNION {{ OPTIONAL {{?s oio:hasRelatedSynonym ?related_synonym}} }} "
        f"UNION {{ OPTIONAL {{?s oio:hasBroadSynonym ?broad_synonym}} }} }} # LIMIT"
    )


def get_most_specific_objects_query(term_iri_list: List[str], predicate: str, ontology: str) -> str:
    return (
        f"SELECT DISTINCT (?cell as ?s) (?cell_label as ?s_label) (?predicate as ?p) (?process as ?o) "
        f"(?function_label as ?o_label)  "
        f"FROM <http://reasoner.renci.org/ontology> FROM <http://reasoner.renci.org/redundant> "
        f"WHERE {{ VALUES ?cell {{ {' '.join(term_iri_list)} }} VALUES ?predicate {{ {predicate} }} "
        f"?cell rdfs:isDefinedBy <{ontology}>. ?cell ?predicate ?process. ?cell rdfs:label ?cell_label. "
        f"?process rdfs:label ?function_label. FILTER NOT EXISTS {{ ?cell ?predicate ?process2. "
        f"?process2 rdfs:subClassOf ?process. FILTER(?process2 != ?process) }} }} # LIMIT"
    )


def get_most_specific_subjects_query(term_iri_list: List[str], predicate: str, ontology: str):
    return (
        f"SELECT DISTINCT (?cell as ?s) (?cell_label as ?s_label) (?predicate as ?p) (?process as ?o) "
        f"(?function_label as ?o_label)  "
        f"FROM <http://reasoner.renci.org/ontology> FROM <http://reasoner.renci.org/redundant> "
        f"WHERE {{ VALUES ?process {{ {' '.join(term_iri_list)} }} VALUES ?predicate {{ {predicate} }} "
        f"?cell rdfs:isDefinedBy <{ontology}>. ?cell ?predicate ?process. ?cell rdfs:label ?cell_label. "
        f"?process rdfs:label ?function_label. FILTER NOT EXISTS {{ ?cell ?predicate ?process2. "
        f"?process2 rdfs:subClassOf ?process. FILTER(?process2 != ?process) }} }} # LIMIT"
    )


def get_obsolete_term_query(seed_list: List[str]) -> str:
    # TODO Add missing implementation. Might not be needed
    raise NotImplementedError


def get_replaced_by_query(term_iri_list: List[str]) -> str:
    # IAO:0100001, term replaced by .
    return (
        f"SELECT * WHERE {{ ?term rdfs:label ?label. ?term owl:deprecated ?depr_status. "
        f"?term IAO:0100001 ?new_term. ?new_term rdfs:label ?new_term_label "
        f"VALUES ?term {{{' '.join(term_iri_list)}}} }}# LIMIT"
    )


# Might not be needed
def get_slim_list_query(ontology: str) -> str:
    return (
        f"SELECT DISTINCT ?slim ?label ?comment WHERE {{ GRAPH ?ontology {{ ?ontology a owl:Ontology. ?ontology "
        f"<http://purl.org/dc/elements/1.1/title> ?title. ?term oio:inSubset ?slim. ?slim rdfs:label ?label. "
        f"?slim rdfs:comment ?comment. FILTER(str(?title) = '{ontology}') }} }}# LIMIT"
    )


def get_slim_members_query(slim_name: str) -> str:
    return (
        f"SELECT ?term WHERE {{ ?term oio:inSubset ?slim. "
        f"?slim rdfs:label ?slim_name. FILTER(str(?slim_name) = '{slim_name}') }}# LIMIT"
    )


def get_ontology_list_query() -> str:
    return (
        "SELECT ?title "
        "WHERE { ?ontology a owl:Ontology. ?ontology <http://purl.org/dc/elements/1.1/title> ?title }# LIMIT"
    )



from typing import List, Optional

from pandasaurus.resources.term import Term


class Slim:
    """Represents upper level slims"""

    def __init__(self, name: str, description: str, term_list: Optional[List[Term]] = None):
        self._name = name
        self._description = description
        self._term_list = term_list

    def get_name(self) -> str:
        """

        Returns:
            str: Slim name

        """
        return self._name

    def get_description(self) -> str:
        """Returns slim description

        Returns:
            Slim description

        """
        return self._description

    def get_term_list(self) -> List[Term]:
        """Returns term list

        Returns:
            Term list

        """
        return self._term_list

    def set_term_list(self, term_list: List[Term]):
        """Sets term list"""
        self._term_list = term_list

    def __str__(self):
        return f"Name:{self._name}, Description: {self._description}"



import logging
import sys


# Create a filter to exclude ERROR log records
class NoErrorFilter(logging.Filter):
    def filter(self, record):
        return record.levelno != logging.ERROR


def configure_logger():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    # logger.propagate = False

    # Create a console handler and set the level to INFO
    info = logging.StreamHandler()
    info.setLevel(logging.INFO)
    # Create a console handler and set the level to ERROR
    error = logging.StreamHandler()
    error.setLevel(logging.ERROR)

    # Create a formatter and set the format for log messages
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    info.setFormatter(formatter)
    error.setFormatter(formatter)
    error.addFilter(NoErrorFilter())
    # Add the console handler to the logger
    logger.addHandler(info)
    logger.addHandler(error)

    if "pytest" not in sys.modules:
        logger.propagate = False

    return logger
